import importlib.machinery
import itertools
import pathlib
import re
import sys
import types
from types import ModuleType
from typing import Optional, Tuple

_module_counter = itertools.count()


ID_REGEX = r"[_A-Za-z][_a-zA-Z0-9]*"
DOTTED_ID_REGEX = rf"{ID_REGEX}(\.{ID_REGEX})*"
ENTRY_POINT_REGEX = rf"^(?P<module>{DOTTED_ID_REGEX})::(?P<func>{ID_REGEX})$"

entry_point = re.compile(ENTRY_POINT_REGEX)


def _parse_entry_point(expr: str) -> Optional[Tuple[str, str]]:
    matches = list(entry_point.finditer(expr))
    if len(matches) != 1:
        return None

    match = matches[0]
    module = match.group("module")
    func = match.group("func")
    assert module is not None
    assert func is not None
    return module, func


def load(path: pathlib.Path, module_name_prefix: str) -> ModuleType:
    if not path.exists() or not path.is_file():
        raise FileNotFoundError(path)

    # NOTE(igarashi): add unique number to module_name to avoid module name conflict
    # NOTE(igarashi): since next(itertools.count) doesn't release GIL, it is an atomic operation
    counter = next(_module_counter)
    module_name = f"pysen._modules.{module_name_prefix}_{counter}"

    loader = importlib.machinery.SourceFileLoader(module_name, str(path))
    module = types.ModuleType(loader.name)
    # NOTE(igarashi): typing raises an error if the module is not included in sys.modules.
    # To allow loaded modules to use dacite, register them to `sys.modules`.
    sys.modules[module_name] = module

    module.__file__ = str(path)
    loader.exec_module(module)

    return module



import difflib
import logging
import pathlib
from abc import abstractmethod
from typing import Iterable, List, Optional, Sequence

from . import git_utils
from .command import CommandBase
from .error_lines import parse_error_diffs
from .reporter import Reporter
from .source import FilePredicateType, Source


class LintCommandBase(CommandBase):
    def __init__(self, base_dir: pathlib.Path, source: Source) -> None:
        self._base_dir = base_dir
        self._source = source

    def _get_sources(
        self, reporter: Reporter, filter_predicate: FilePredicateType
    ) -> List[pathlib.Path]:
        sources = self.source.resolve_files(
            self.base_dir,
            filter_predicate,
            self.git_enabled(),
            reporter,
        )
        return sorted(sources)

    def _get_covered_files(
        self,
        reporter: Reporter,
        files: Sequence[pathlib.Path],
        filter_predicate: FilePredicateType,
    ) -> List[pathlib.Path]:
        sources = self._get_sources(reporter, filter_predicate)
        covered: List[pathlib.Path] = []

        for f in files:
            if f in sources:
                covered.append(f)
            else:
                reporter.logger.info(f"Skipping {f} for {self.name}")

        return covered

    @property
    def base_dir(self) -> pathlib.Path:
        return self._base_dir

    def git_enabled(self) -> bool:
        return git_utils.check_git_available(self.base_dir)

    @property
    def source(self) -> Source:
        return self._source


class SingleFileLintCommandBase(LintCommandBase):
    def _run(self, reporter: Reporter, file_paths: Iterable[pathlib.Path]) -> int:
        # NOTE(igarashi): create a list to evaluate check() for all file paths
        if all([self.check(file_path, reporter) for file_path in file_paths]):
            return 0
        else:
            return 1

    def __call__(self, reporter: Reporter) -> int:
        file_paths = self._get_sources(reporter, self.filter)
        reporter.logger.info(f"Checking {len(file_paths)} files")
        return self._run(reporter, file_paths)

    def run_files(self, reporter: Reporter, files: Sequence[pathlib.Path]) -> int:
        covered_files = self._get_covered_files(reporter, files, self.filter)
        return self._run(reporter, covered_files)

    @abstractmethod
    def filter(self, file_path: pathlib.Path) -> bool:
        ...

    @abstractmethod
    def check(self, file_path: pathlib.Path, reporter: Reporter) -> bool:
        ...


class SingleFileFormatCommandBase(SingleFileLintCommandBase):
    def __init__(
        self, base_dir: pathlib.Path, source: Source, inplace_edit: bool
    ) -> None:
        super().__init__(base_dir, source)
        self._inplace_edit = inplace_edit

    @property
    def inplace_edit(self) -> bool:
        return self._inplace_edit

    def check(self, file_path: pathlib.Path, reporter: Reporter) -> bool:
        formatted = self.format(file_path, reporter)
        if formatted is None:
            return False

        if self._inplace_edit:
            with file_path.open(mode="w") as f:
                f.write(formatted)
            return True

        else:
            with file_path.open() as f:
                original = f.readlines()
            diff = "".join(
                difflib.unified_diff(
                    original,
                    formatted.splitlines(True),
                    fromfile=str(file_path),
                    tofile=str(file_path),
                )
            )
            if len(diff) == 0:
                return True
            else:
                reporter.process_output.log(logging.INFO, diff)
                diagnostics = parse_error_diffs(
                    diff, lambda _: file_path, logger=reporter.logger
                )
                reporter.report_diagnostics(list(diagnostics))
                return False

    @abstractmethod
    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        """Returns formatted content without modifying the original file.
        Note:
            If a file cannot be formatted due to its content (e.g. invalid syntax),
            this method should return `None`.
            In this case, `SingleFileFormatCommandBase` continues to check other files.
        """
        ...



import copy
import dataclasses
import functools
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple

from pysen import process_utils
from pysen.command import check_command_installed
from pysen.dist_version import get_version
from pysen.error_lines import parse_error_lines
from pysen.exceptions import IncompatibleVersionError
from pysen.path import change_dir
from pysen.py_version import VersionRepresentation
from pysen.reporter import Reporter
from pysen.setting import SettingBase, to_dash_case

_SettingFileName = "setup.cfg"


def _contains(target: Sequence[str], item: str) -> bool:
    code_category = item[0].upper()

    for x in target:
        if len(x) == 1:
            if x.upper() == code_category:
                return True
        else:
            if x == item:
                return True

    return False


@dataclasses.dataclass
class Flake8Setting(SettingBase):
    max_line_length: int = 88
    select: Optional[List[str]] = None
    ignore: Optional[List[str]] = None
    enable_extensions: Optional[List[str]] = None
    max_complexity: Optional[int] = None

    _comments: List[str] = dataclasses.field(default_factory=list)

    @staticmethod
    def default() -> "Flake8Setting":
        return Flake8Setting(
            select=["B", "C", "E", "F", "W", "B950"],
        ).to_black_compatible()

    def to_black_compatible(self) -> "Flake8Setting":
        new = copy.deepcopy(self)
        if new.ignore is None:
            new.ignore = []

        if not _contains(new.ignore, "E203"):
            new.ignore.append("E203")
            new._comments.append("# E203: black treats : as a binary operator")

        if not _contains(new.ignore, "E231"):
            new.ignore.append("E231")
            new._comments.append("# E231: black doesn't put a space after ,")

        if not _contains(new.ignore, "E501"):
            new.ignore.append("E501")
            new._comments.append(
                "# E501: black may exceed the line-length to follow other style rules"
            )

        if not _contains(new.ignore, "E701"):
            new.ignore.append("E701")
            new._comments.append(
                "# E701: black will collapse ... only functions etc. to a single line"
            )

        if not _contains(new.ignore, "E704"):
            new.ignore.append("E704")
            new._comments.append(
                "# E704: black will collapse ... only functions etc. to a single line"
            )

        W503_or_504_enabled = _contains(new.ignore, "W503") or _contains(
            new.ignore, "W504"
        )
        if not W503_or_504_enabled:
            new.ignore.append("W503")
            new._comments.append(
                "# W503 or W504: either one needs to be disabled to select W error codes"
            )

        return new

    def export(self) -> Tuple[Sequence[str], Dict[str, Any]]:
        section_name = "flake8"
        entries = self.asdict(
            omit_none=True, naming_rule=to_dash_case, ignore_fields=["_comments"]
        )
        for c in self._comments:
            assert c.startswith("#") and c not in entries
            entries[c] = None

        return [section_name], entries


@functools.lru_cache(1)
def _check_flake8_version() -> None:
    version = get_version("flake8")
    minimum_supported = VersionRepresentation(3, 8)
    if version < minimum_supported:
        raise IncompatibleVersionError(
            f"pysen only supports flake8 >= {minimum_supported}, "
            f"version {version} is not supported."
        )


def run(
    reporter: Reporter,
    base_dir: pathlib.Path,
    setting_path: pathlib.Path,
    sources: Iterable[pathlib.Path],
) -> int:
    check_command_installed(*process_utils.add_python_executable("flake8", "--version"))
    _check_flake8_version()
    targets = [str(d) for d in sources]
    if len(targets) == 0:
        return 0

    cmd = ["flake8", "--config", str(setting_path)] + targets
    with change_dir(base_dir):
        ret, stdout, _ = process_utils.run(
            process_utils.add_python_executable(*cmd), reporter
        )

    diagnostics = parse_error_lines(stdout, logger=reporter.logger)
    reporter.report_diagnostics(list(diagnostics))

    return ret



import dataclasses
import pathlib
from typing import Any, Dict, Optional

import dacite
import tomlkit

from .exceptions import InvalidConfigurationError
from .pyproject_model import _get_descendant, _workaround_tomlkit_unmarshal


@dataclasses.dataclass
class CliConfig:
    settings_dir: Optional[pathlib.Path] = None


def _expand_path(base_dir: pathlib.Path, s: Any) -> pathlib.Path:
    if isinstance(s, pathlib.Path):
        return s
    elif isinstance(s, str):
        return base_dir / s
    else:
        raise dacite.WrongTypeError(pathlib.Path, s)


def _load_cli_section(path: pathlib.Path) -> Optional[Dict[str, Any]]:
    with path.open("r") as f:
        pyproject = tomlkit.loads(f.read())

    section = _get_descendant(pyproject, ["tool", "pysen-cli"])
    if section is None:
        return None

    data = _workaround_tomlkit_unmarshal(section)
    assert isinstance(data, dict)
    return data


def _parse_dict(data: Dict[str, Any], base_dir: pathlib.Path) -> CliConfig:
    dacite_config = dacite.Config(
        type_hooks={pathlib.Path: lambda x: _expand_path(base_dir, x)},
        strict=True,
    )

    try:
        config = dacite.from_dict(CliConfig, data, dacite_config)
        assert isinstance(config, CliConfig)
        return config
    except dacite.DaciteError as e:
        raise InvalidConfigurationError(f"invalid configuration: {e}") from None


def parse(path: pathlib.Path) -> Optional[CliConfig]:
    base_dir = path.resolve().parent
    section = _load_cli_section(path)
    if section is None:
        return None
    return _parse_dict(section, base_dir)



import dataclasses
import logging
import pathlib
from typing import Any, Dict, List, Optional, Sequence

import dacite
import tomlkit

from ._version import __version__
from .exceptions import InvalidConfigurationError, PysenSectionNotFoundError
from .factory import ConfigureLintOptions, MypyModuleOption
from .isort import IsortSectionName
from .mypy import MypyFollowImports, MypyPreset, MypyTarget
from .py_version import PythonVersion, VersionRepresentation
from .source import Source

_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class LintConfig(ConfigureLintOptions):
    base: Optional[pathlib.Path] = None

    # * Alias Fields *
    # config to create mypy_modules instances
    mypy_ignore_packages: Optional[List[str]] = None

    # * Deprecated Fields *
    mypy_target_dirs: Optional[List[pathlib.Path]] = None

    def update(self, rhs: "LintConfig") -> None:
        # NOTE(igarashi): do not overwrite base by rhs.base as it is required to refer
        fields = dataclasses.fields(self)
        for field in fields:
            field_name = field.name
            if field_name == "base":
                continue

            # NOTE(igarashi): handle None in rhs as "unset" state
            value = getattr(rhs, field_name)
            if value is not None:
                setattr(self, field_name, value)


@dataclasses.dataclass
class PluginConfig:
    # NOTE(igarashi): `location` is set from _parse_plugin_configs, not from pyproject.toml
    location: str = ""
    function: Optional[str] = None
    script: Optional[pathlib.Path] = None
    config: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class Config:
    version: Optional[VersionRepresentation] = None
    lint: Optional[LintConfig] = None
    builder: Optional[pathlib.Path] = None
    plugin: Optional[List[PluginConfig]] = None


PLUGIN_PATH_ROOT = "tool.pysen.plugin"


def _parse_plugin_configs(base_dir: pathlib.Path, data: Any) -> List[PluginConfig]:
    if data is None:
        return []

    if not isinstance(data, dict):
        raise dacite.WrongTypeError(dict, data, PLUGIN_PATH_ROOT)

    result: List[PluginConfig] = []

    for key, value in data.items():
        path = f"{PLUGIN_PATH_ROOT}.{key}"

        if not isinstance(value, dict):
            raise dacite.WrongTypeError(dict, value, path)

        # NOTE(igarashi): ensure that `location` is unspecified in pyproject.toml
        if "location" in value:
            raise dacite.DaciteError(f"unknown filed: {path}.location")

        config = dacite.from_dict(
            PluginConfig,
            value,
            config=dacite.Config(
                type_hooks={pathlib.Path: lambda x: _expand_path(base_dir, x)},
                strict=True,
            ),
        )
        assert isinstance(config, PluginConfig)
        config.location = path

        if config.function is None and config.script is None:
            raise dacite.DaciteError(
                "must specify either function or script for plugin field"
            )

        if config.function is not None and config.script is not None:
            raise dacite.DaciteError("only one of function or script must be specified")

        result.append(config)

    return result


def _parse_mypy_follow_imports(s: Any) -> MypyFollowImports:
    if not isinstance(s, str):
        raise dacite.WrongTypeError(MypyFollowImports, s)

    try:
        return MypyFollowImports[s.upper()]
    except KeyError:
        raise dacite.DaciteError(f"invalid follow_imports value: {s}") from None


def _parse_mypy_preset(s: Any) -> MypyPreset:
    if not isinstance(s, str):
        raise dacite.WrongTypeError(MypyPreset, s)

    try:
        return MypyPreset[s.upper()]
    except KeyError:
        raise dacite.DaciteError(f"invalid mypy_preset value: {s}") from None


def _parse_isort_section_name(s: Any) -> IsortSectionName:
    if not isinstance(s, str):
        raise dacite.WrongTypeError(MypyPreset, s)

    try:
        return IsortSectionName[s.upper()]
    except KeyError:
        raise dacite.DaciteError(f"invalid default_section value: {s}") from None


def _parse_python_version(s: Any) -> PythonVersion:
    if not isinstance(s, str):
        raise dacite.WrongTypeError(str, s)

    try:
        return PythonVersion.parse_short_representation(s)
    except KeyError as e:
        raise dacite.DaciteError(str(e))


def _expand_path(base_dir: pathlib.Path, s: Any) -> pathlib.Path:
    if isinstance(s, pathlib.Path):
        return s
    elif isinstance(s, str):
        return base_dir / s
    else:
        raise dacite.WrongTypeError(pathlib.Path, s)


def _parse_source(base_dir: pathlib.Path, d: Any) -> Source:
    if isinstance(d, list):
        return Source(includes=[_expand_path(base_dir, x) for x in d])
    elif isinstance(d, dict):

        @dataclasses.dataclass
        class _SourceConfig:
            includes: Optional[List[pathlib.Path]] = None
            include_globs: Optional[List[str]] = None
            excludes: Optional[List[pathlib.Path]] = None
            exclude_globs: Optional[List[str]] = None

        config = dacite.from_dict(
            _SourceConfig,
            d,
            config=dacite.Config(
                type_hooks={pathlib.Path: lambda x: _expand_path(base_dir, x)},
                strict=True,
            ),
        )
        source = Source(includes=config.includes, excludes=config.excludes)
        if config.include_globs is not None:
            for i in config.include_globs:
                source.add_include(i, glob=True, base_dir=base_dir)

        if config.exclude_globs is not None:
            for e in config.exclude_globs:
                source.add_exclude(e, glob=True, base_dir=base_dir)

        if len(source.includes) == 0:
            source.add_include(".", base_dir=base_dir)

        return source

    else:
        raise dacite.DaciteError(f"invalid source value: {d}") from None


def _parse_mypy_target(base_dir: pathlib.Path, d: Any) -> MypyTarget:
    if not isinstance(d, dict):
        raise dacite.WrongTypeError(dict, d, "tool.pysen.lint.mypy_targets")

    target = dacite.from_dict(
        MypyTarget,
        d,
        config=dacite.Config(
            type_hooks={pathlib.Path: lambda x: _expand_path(base_dir, x)},
            strict=True,
        ),
    )
    assert isinstance(target, MypyTarget)
    if len(target.paths) == 0:
        raise dacite.DaciteError(
            "invalid mypy_target: each target must have one or more paths"
        )

    return target


def _parse_mypy_targets(base_dir: pathlib.Path, config: Any) -> List[MypyTarget]:
    if not isinstance(config, list):
        raise dacite.WrongTypeError(List[MypyTarget], config)

    if not all(isinstance(x, dict) for x in config):
        raise dacite.DaciteError("tool.pysen.lint.mypy_targets must be a list of dicts")

    return [_parse_mypy_target(base_dir, x) for x in config]


def _parse_mypy_modules(config: Any) -> Dict[str, MypyModuleOption]:
    if not isinstance(config, dict):
        raise dacite.WrongTypeError(Dict[str, MypyModuleOption], config)

    mypy_modules: Dict[str, MypyModuleOption] = {}

    for target_module, option_dict in config.items():
        if not isinstance(target_module, str):
            raise dacite.WrongTypeError(
                str, target_module, "tool.pysen.lint.mypy_modules"
            )

        if not isinstance(option_dict, dict):
            raise dacite.WrongTypeError(
                MypyModuleOption,
                option_dict,
                f'tool.pysen.lint.mypy_modules."{target_module}"',
            )

        try:
            module_option = dacite.from_dict(
                MypyModuleOption,
                option_dict,
                config=dacite.Config(
                    strict=True,
                    type_hooks={
                        MypyPreset: _parse_mypy_preset,
                        MypyFollowImports: _parse_mypy_follow_imports,
                    },
                ),
            )
            assert isinstance(module_option, MypyModuleOption)
            mypy_modules[target_module] = module_option
        except ValueError as e:
            raise dacite.DaciteError(f"invalid mypy_module: {target_module}, {e}")

    return mypy_modules


def _parse_dict(data: Dict[str, Any], base_dir: pathlib.Path) -> Config:
    dacite_config = dacite.Config(
        type_hooks={
            MypyPreset: _parse_mypy_preset,
            IsortSectionName: _parse_isort_section_name,
            pathlib.Path: lambda x: _expand_path(base_dir, x),
            Source: lambda x: _parse_source(base_dir, x),
            List[PluginConfig]: lambda x: _parse_plugin_configs(base_dir, x),
            VersionRepresentation: VersionRepresentation.from_str,
            PythonVersion: _parse_python_version,
            List[MypyTarget]: lambda x: _parse_mypy_targets(base_dir, x),
            Dict[str, MypyModuleOption]: _parse_mypy_modules,
        },
        strict=True,
    )

    try:
        config = dacite.from_dict(Config, data, dacite_config)
        assert isinstance(config, Config)
        return config
    except dacite.DaciteError as e:
        raise InvalidConfigurationError(f"invalid configuration: {e}") from None


def _workaround_tomlkit_unmarshal(data: Any) -> Any:
    if data is None or isinstance(data, tomlkit.items.Null):
        return None
    elif isinstance(data, dict):
        # tomlkit.items.Dict, tomlkit.container.Container
        ret: Dict[str, Any] = {}
        for k, v in data.items():
            k = _workaround_tomlkit_unmarshal(k)
            v = _workaround_tomlkit_unmarshal(v)
            ret[k] = v

        return ret
    elif isinstance(data, list):
        # tomlkit.items.Array
        return list([_workaround_tomlkit_unmarshal(v) for v in data])
    elif isinstance(data, tomlkit.items.Bool):
        return bool(data)
    elif isinstance(data, tomlkit.items.Float):
        return float(data)
    elif isinstance(data, tomlkit.items.Integer):
        return int(data)
    elif isinstance(data, tomlkit.items.String):
        return str(data)
    elif isinstance(
        data, (tomlkit.items.DateTime, tomlkit.items.Date, tomlkit.items.Time)
    ):
        raise NotImplementedError(f"tomlkit type: {type(data)}")

    return data


def _get_descendant(
    document: tomlkit.toml_document.TOMLDocument, paths: Sequence[str]
) -> Optional[Any]:
    assert len(paths) > 0

    node: Dict[str, Any] = document
    for p in paths[:-1]:
        e = node.get(p)
        if not isinstance(e, dict):
            return None
        node = e

    return node.get(paths[-1])


def get_tool_section(
    tool_name: str,
    pyproject: tomlkit.toml_document.TOMLDocument,
) -> Optional[tomlkit.toml_document.TOMLDocument]:
    return _get_descendant(pyproject, ["tool", tool_name])


def has_tool_section(
    tool_name: str,
    pyproject: tomlkit.toml_document.TOMLDocument,
) -> bool:
    return get_tool_section(tool_name, pyproject) is not None


def _load_pysen_section(path: pathlib.Path) -> Dict[str, Any]:
    if not path.exists():
        raise FileNotFoundError(path)

    with path.open("r") as f:
        pyproject = tomlkit.loads(f.read())

    section = get_tool_section("pysen", pyproject)
    if section is None:
        section = get_tool_section("jiro", pyproject)
        if section is not None:
            _logger.warning(
                "jiro section under a config file is deprecated. Use pysen instead."
            )

    if section is None:
        raise PysenSectionNotFoundError(str(path))

    data = _workaround_tomlkit_unmarshal(section)
    assert isinstance(data, dict)
    return data


def _migrate_alias_fields(config: Config) -> None:
    if config.lint is not None:
        lint_config = config.lint

        if lint_config.mypy_ignore_packages is not None:
            mypy_modules: Dict[str, MypyModuleOption] = lint_config.mypy_modules or {}

            for m in lint_config.mypy_ignore_packages:
                if m in mypy_modules:
                    raise dacite.DaciteError(
                        f"{m} is configured in both mypy_ignore_packages and mypy_modules"
                    )

                mypy_modules[m] = MypyModuleOption(
                    ignore_errors=True, follow_imports=MypyFollowImports.SKIP
                )

            lint_config.mypy_ignore_packages = None
            lint_config.mypy_modules = mypy_modules


def _migrate_deprecated_fields(path: pathlib.Path, config: Config) -> None:
    if config.lint is not None:
        if config.lint.mypy_target_dirs is not None:
            _logger.warning(
                "tool.pysen.mypy_target_dirs is deprecated since 0.6.0, "
                f"Use tool.pysen.mypy_targets instead (File: {path})"
            )
            config.lint.mypy_targets = [MypyTarget(config.lint.mypy_target_dirs)]
            config.lint.mypy_target_dirs = None


def _load_version(data: Dict[str, Any]) -> Optional[VersionRepresentation]:
    version = data.get("version")
    if version is None:
        return None

    try:
        return VersionRepresentation.from_str(version)
    except ValueError as e:
        raise InvalidConfigurationError(e) from None


def _check_version(
    file_path: pathlib.Path,
    config_version: Optional[VersionRepresentation],
    actual_version: VersionRepresentation,
) -> None:
    if config_version is None:
        _logger.warning(
            "Consider specifying 'version' under [tool.pysen] section in your config "
            "to check compliance against the version of the installed pysen. "
            f"(File: {file_path})"
        )
    elif not config_version.is_compatible(actual_version):
        _logger.warning(
            f"config specifies version {config_version}, "
            f"but the pysen you are using is version {actual_version}, "
            "which might not be compatible. "
            f"(File: {file_path})"
        )


def parse(path: pathlib.Path) -> Config:
    path = path.resolve()
    base_dir = path.parent

    section = _load_pysen_section(path)
    version = _load_version(section)
    _check_version(path, version, VersionRepresentation.from_str(__version__))

    config = _parse_dict(section, base_dir)
    _migrate_alias_fields(config)
    _migrate_deprecated_fields(path, config)
    return config



import dataclasses
import enum
import functools
import pathlib
from typing import Any, Dict, List, Optional, Sequence, Tuple

from pysen import process_utils
from pysen.command import check_command_installed
from pysen.dist_version import get_version
from pysen.error_lines import parse_error_lines
from pysen.exceptions import IncompatibleVersionError
from pysen.path import PathLikeType, change_dir, get_relative_path, resolve_path
from pysen.py_version import PythonVersion
from pysen.reporter import Reporter
from pysen.setting import SettingBase

_IgnoreFields: List[str] = ["_pysen_convert_abspath"]


class MypyFollowImports(enum.Enum):
    NORMAL = "normal"
    SILENT = "silent"
    SKIP = "skip"
    ERROR = "error"


@dataclasses.dataclass
class MypyPlugin:
    script: Optional[pathlib.Path] = None
    function: Optional[str] = None

    def __post_init__(self) -> None:
        if self.script is None and self.function is None:
            raise ValueError("either script or function must be specified")

        if self.script is not None and self.function is not None:
            raise ValueError("cannot specify both script and function")

    def as_config(self, relative_from: Optional[pathlib.Path] = None) -> str:
        if self.function is not None:
            return self.function

        assert self.script is not None
        if relative_from is not None:
            return get_relative_path(self.script, relative_from)
        else:
            return str(self.script)


@dataclasses.dataclass
class MypySetting(SettingBase):
    python_version: Optional[PythonVersion] = None

    check_untyped_defs: Optional[bool] = None
    disallow_any_decorated: Optional[bool] = None
    disallow_any_generics: Optional[bool] = None
    disallow_any_unimported: Optional[bool] = None
    disallow_incomplete_defs: Optional[bool] = None
    disallow_subclassing_any: Optional[bool] = None
    disallow_untyped_calls: Optional[bool] = None
    disallow_untyped_decorators: Optional[bool] = None
    disallow_untyped_defs: Optional[bool] = None
    follow_imports: Optional[MypyFollowImports] = None
    ignore_errors: Optional[bool] = None
    ignore_missing_imports: Optional[bool] = None
    mypy_path: Optional[List[PathLikeType]] = None
    no_implicit_optional: Optional[bool] = None
    pretty: Optional[bool] = None
    show_error_codes: Optional[bool] = None
    strict_equality: Optional[bool] = None
    strict_optional: Optional[bool] = None
    warn_redundant_casts: Optional[bool] = None
    warn_return_any: Optional[bool] = None
    warn_unreachable: Optional[bool] = None
    warn_unused_configs: Optional[bool] = None
    warn_unused_ignores: Optional[bool] = None
    plugins: Optional[List[MypyPlugin]] = None

    # configuration for export settings
    _pysen_convert_abspath: bool = False

    @staticmethod
    def _top(**kwargs: Any) -> "MypySetting":
        updates = {
            "check_untyped_defs": True,
            "disallow_any_decorated": True,
            "disallow_any_generics": True,
            "disallow_any_unimported": True,
            "disallow_incomplete_defs": True,
            "disallow_subclassing_any": True,
            "disallow_untyped_calls": True,
            "disallow_untyped_decorators": True,
            "disallow_untyped_defs": True,
            "ignore_errors": False,
            "ignore_missing_imports": False,
            "no_implicit_optional": True,
            "show_error_codes": True,
            "strict_equality": True,
            "strict_optional": True,
            "warn_redundant_casts": True,
            "warn_return_any": True,
            "warn_unreachable": True,
            "warn_unused_configs": True,
            "warn_unused_ignores": True,
        }
        updates.update(kwargs)
        return MypySetting(**updates)  # type: ignore

    @staticmethod
    def very_strict(**kwargs: Any) -> "MypySetting":
        updates = {
            "disallow_any_decorated": False,
            "disallow_any_unimported": False,
            "disallow_untyped_decorators": False,
            "ignore_missing_imports": True,
        }
        updates.update(kwargs)
        setting = MypySetting._top(**updates)
        return setting

    @staticmethod
    def strict(**kwargs: Any) -> "MypySetting":
        updates = {
            "warn_unused_ignores": False,
            "disallow_any_generics": False,
        }
        updates.update(kwargs)
        setting = MypySetting.very_strict(**updates)
        return setting

    @staticmethod
    def entry(**kwargs: Any) -> "MypySetting":
        updates = {
            "disallow_untyped_calls": False,
            "disallow_untyped_defs": False,
            "warn_return_any": False,
        }
        updates.update(kwargs)
        setting = MypySetting.strict(**updates)
        return setting

    def export(
        self, base_dir: pathlib.Path, target_module: Optional[str] = None
    ) -> Tuple[Sequence[str], Dict[str, Any]]:
        section_name = "mypy"
        if target_module is not None:
            section_name += f"-{target_module}"

        # TODO(igarashi): refactor these flaky code, see: black.py
        entries = self.asdict(
            _IgnoreFields
            + ["python_version", "mypy_path", "plugins", "follow_imports"],
            omit_none=True,
        )
        if self.python_version is not None:
            entries["python_version"] = self.python_version.version

        if self.mypy_path is not None:
            mypy_path: List[str] = []
            for p in self.mypy_path:
                path: str
                if self._pysen_convert_abspath:
                    path = get_relative_path(p, base_dir)
                else:
                    path = str(p)

                mypy_path.append(path)

            entries["mypy_path"] = mypy_path

        if self.plugins is not None:
            relative_from: Optional[pathlib.Path] = None
            if self._pysen_convert_abspath:
                relative_from = base_dir

            entries["plugins"] = [p.as_config(relative_from) for p in self.plugins]

        if self.follow_imports is not None:
            entries["follow_imports"] = self.follow_imports.value

        return [section_name], entries


@dataclasses.dataclass
class MypyTarget:
    paths: List[pathlib.Path]
    namespace_packages: bool = False


@functools.lru_cache(1)
def _check_mypy_version() -> None:
    version = get_version("mypy")
    if version.major >= 2 or version.major == 0 and version.minor < 770:
        raise IncompatibleVersionError(
            f"pysen only supports mypy version >=0.770, <2. "
            f"version {version} is not supported."
        )


def run(
    reporter: Reporter,
    base_dir: pathlib.Path,
    setting_path: pathlib.Path,
    target: MypyTarget,
    require_diagnostics: bool,
) -> int:
    check_command_installed(*process_utils.add_python_executable("mypy", "--version"))
    _check_mypy_version()

    target_paths = [str(resolve_path(base_dir, x)) for x in target.paths]
    if len(target_paths) == 0:
        return 0

    extra_options: List[str] = ["--show-absolute-path"]
    if require_diagnostics:
        extra_options += [
            "--no-color-output",
            "--show-column-numbers",
            "--no-error-summary",
        ]
    else:
        extra_options += [
            "--pretty",
        ]

    if target.namespace_packages:
        extra_options.append("--namespace-packages")

    cmd = ["mypy"] + extra_options + ["--config-file", str(setting_path)] + target_paths
    with change_dir(base_dir):
        ret, stdout, _ = process_utils.run(
            process_utils.add_python_executable(*cmd), reporter
        )

    if require_diagnostics:
        diagnostics = parse_error_lines(stdout, logger=reporter.logger)
        reporter.report_diagnostics(list(diagnostics))

    return ret



import pathlib
from typing import Optional, Sequence

from . import py_module
from .component import ComponentBase
from .exceptions import InvalidManifestBuilderError
from .manifest import Manifest, ManifestBase


def _build_external(
    path: pathlib.Path,
    components: Sequence[ComponentBase],
    src_path: Optional[pathlib.Path],
) -> ManifestBase:
    module = py_module.load(path, "builder")
    entry_point = getattr(module, "build", None)
    if entry_point is None or not callable(entry_point):
        raise InvalidManifestBuilderError(
            path, "external builder must have `build` method"
        )

    ret = entry_point(components, src_path)
    if not isinstance(ret, ManifestBase):
        raise InvalidManifestBuilderError(
            path, "`build` must return an instance of ManifestBase"
        )

    return ret


def _build(
    components: Sequence[ComponentBase], src_path: Optional[pathlib.Path]
) -> ManifestBase:
    return Manifest(components)


def build(
    components: Sequence[ComponentBase],
    src_path: Optional[pathlib.Path] = None,
    external_builder: Optional[pathlib.Path] = None,
) -> ManifestBase:
    if external_builder is not None:
        return _build_external(external_builder, components, src_path)

    return _build(components, src_path)



import functools
import logging
import os
import pathlib
import threading
from typing import Sequence, Tuple, cast

_logger = logging.getLogger(__name__)
_lock = threading.Lock()

try:
    import git
    from git import Blob  # type: ignore

    # NOTE: Upstream issue: https://github.com/gitpython-developers/GitPython/issues/1349

    _git_available = True
except ImportError:
    _git_available = False
    _logger.warning("[pysen.git_utils] git is not available")


class GitRepositoryNotFoundError(Exception):
    pass


def _check_git_enabled() -> bool:
    if not _git_available:
        return False

    if os.environ.get("PYSEN_IGNORE_GIT", "0") != "0":
        return False

    return True


def check_git_available(target_dir: pathlib.Path) -> bool:
    with _lock:
        try:
            if not _check_git_enabled():
                return False

            with git.Repo(target_dir, search_parent_directories=True):
                return True
        except git.InvalidGitRepositoryError:
            return False


def _list_indexed_files(target_dir: pathlib.Path) -> Sequence[pathlib.Path]:
    if not _check_git_enabled():
        return []

    # Ensure abs_target_dir ends with /
    # We avoid pathlib.Path because the loop calling predicate is performance critical.
    abs_target_dir = os.path.join(str(target_dir.resolve()), "")

    def predicate(item: Tuple[int, Blob]) -> bool:
        blob = item[1]
        ret: bool = blob.abspath.startswith(abs_target_dir)
        return ret

    try:
        with git.Repo(target_dir, search_parent_directories=True) as repo:
            deleted_files = set(
                diff.a_blob.abspath
                for diff in repo.index.diff(None)
                if diff.change_type == "D"
            )
            blobs = set(blob.abspath for _, blob in repo.index.iter_blobs(predicate))
            return [pathlib.Path(abspath) for abspath in blobs - deleted_files]
    except git.InvalidGitRepositoryError:
        raise GitRepositoryNotFoundError() from None


@functools.lru_cache(8)
def list_indexed_files(target_dir: pathlib.Path) -> Sequence[pathlib.Path]:
    with _lock:
        return _list_indexed_files(target_dir)


def _check_tracked(path: pathlib.Path) -> bool:
    if not _check_git_enabled():
        return False

    # TODO(igarashi) use git command directly for better performance
    abspath = str(path.expanduser().resolve())

    def predicate(item: Tuple[int, Blob]) -> bool:
        blob = item[1]
        return cast(bool, blob.abspath == abspath)

    try:
        with git.Repo(path, search_parent_directories=True) as repo:
            items = list(repo.index.iter_blobs(predicate))
            return len(items) > 0
    except git.InvalidGitRepositoryError:
        raise GitRepositoryNotFoundError() from None


def check_tracked(path: pathlib.Path) -> bool:
    with _lock:
        return _check_tracked(path)



import dataclasses
import pathlib
from typing import Dict, List, Optional

from .black import Black, BlackSetting
from .component import ComponentBase
from .flake8 import Flake8, Flake8Setting
from .isort import Isort, IsortSectionName, IsortSetting
from .mypy import (
    Mypy,
    MypyFollowImports,
    MypyPlugin,
    MypyPreset,
    MypySetting,
    MypyTarget,
)
from .py_version import PythonVersion
from .source import Source


@dataclasses.dataclass
class MypyModuleOption:
    preset: Optional[MypyPreset] = None
    ignore_errors: bool = False
    follow_imports: Optional[MypyFollowImports] = None

    def __post_init__(self) -> None:
        if self.preset is not None and self.ignore_errors:
            raise ValueError("cannot specify both preset and ignore_errors")

    def get_setting(self) -> MypySetting:
        if self.ignore_errors:
            return MypySetting(ignore_errors=True, follow_imports=self.follow_imports)

        preset: MypyPreset
        if self.preset is not None:
            preset = self.preset
        else:
            preset = MypyPreset.STRICT

        return preset.get_setting(follow_imports=self.follow_imports)


@dataclasses.dataclass
class ConfigureLintOptions:
    enable_black: Optional[bool] = None
    enable_flake8: Optional[bool] = None
    enable_isort: Optional[bool] = None
    enable_mypy: Optional[bool] = None
    mypy_preset: Optional[MypyPreset] = None
    mypy_modules: Optional[Dict[str, MypyModuleOption]] = None
    source: Optional[Source] = None
    line_length: Optional[int] = None
    py_version: Optional[PythonVersion] = None
    isort_known_third_party: Optional[List[str]] = None
    isort_known_first_party: Optional[List[str]] = None
    isort_default_section: Optional[IsortSectionName] = None
    mypy_path: Optional[List[pathlib.Path]] = None
    mypy_plugins: Optional[List[MypyPlugin]] = None
    mypy_targets: Optional[List[MypyTarget]] = None


def configure_lint(options: ConfigureLintOptions) -> List[ComponentBase]:
    components: List[ComponentBase] = []

    python_version: PythonVersion
    if options.py_version is not None:
        python_version = options.py_version
    else:
        python_version = PythonVersion(3, 8)

    line_length = options.line_length or 88

    # NOTE: `isort` may format code in a way that violates `black` rules
    # Apply `isort` after `black` to avoid such violation
    if options.enable_isort:
        isort_setting = IsortSetting.default()
        isort_setting.line_length = line_length
        isort_setting.default_section = (
            options.isort_default_section or IsortSectionName.THIRDPARTY
        )
        if options.isort_known_third_party is not None:
            isort_setting.known_third_party = set(options.isort_known_third_party)
        if options.isort_known_first_party is not None:
            isort_setting.known_first_party = set(options.isort_known_first_party)

        if options.enable_black:
            isort_setting = isort_setting.to_black_compatible()

        isort = Isort(setting=isort_setting, source=options.source)
        components.append(isort)

    if options.enable_black:
        black_setting = BlackSetting.default(python_version)
        black_setting.line_length = line_length
        black = Black(setting=black_setting, source=options.source)
        components.append(black)

    if options.enable_flake8:
        flake8_setting = Flake8Setting.default()
        flake8_setting.max_line_length = line_length
        if options.enable_black:
            flake8_setting = flake8_setting.to_black_compatible()

        flake8 = Flake8(setting=flake8_setting, source=options.source)
        components.append(flake8)

    if options.enable_mypy:
        if options.mypy_preset is not None:
            mypy_setting = options.mypy_preset.get_setting()
        else:
            mypy_setting = MypySetting.strict()
        mypy_setting.python_version = python_version
        if options.mypy_path is not None:
            mypy_setting.mypy_path = list(options.mypy_path)
        if options.mypy_plugins is not None:
            mypy_setting.plugins = list(options.mypy_plugins)

        mypy_module_settings: Dict[str, MypySetting] = {}
        if options.mypy_modules is not None:
            for module_name, module_option in options.mypy_modules.items():
                mypy_module_settings[module_name] = module_option.get_setting()

        mypy = Mypy(
            setting=mypy_setting,
            module_settings=mypy_module_settings,
            mypy_targets=options.mypy_targets,
        )
        components.append(mypy)

    return components



import dataclasses
import pathlib


@dataclasses.dataclass(frozen=True)
class PathContext:
    base_dir: pathlib.Path
    settings_dir: pathlib.Path


@dataclasses.dataclass(frozen=True)
class RunOptions:
    require_diagnostics: bool = True
    no_parallel: bool = False



import dataclasses
import re
from typing import Optional


def _coerce_type(s: Optional[str]) -> Optional[int]:
    if s is None:
        return None
    else:
        return int(s)


@dataclasses.dataclass(frozen=True)
class VersionRepresentation:
    major: int
    minor: int
    patch: Optional[int] = None
    pre_release: Optional[str] = None

    @property
    def version(self) -> str:
        s = f"{self.major}.{self.minor}"
        if self.patch is not None:
            s += f".{self.patch}"

        return s

    @classmethod
    def from_str(cls, s: str) -> "VersionRepresentation":
        number = r"(?:0|[1-9]\d*)"
        major = rf"^(?P<major>{number})"
        minor = rf"\.(?P<minor>{number})"
        patch = rf"(\.(?P<patch>{number}))?"
        pre_release = rf"(?P<pre_release>(a|b|rc){number})?$"
        pattern = major + minor + patch + pre_release
        m = re.match(pattern, s)

        if m is not None:
            return cls(
                major=int(m.group("major")),
                minor=int(m.group("minor")),
                patch=_coerce_type(m.group("patch")),
                pre_release=m.group("pre_release"),
            )
        raise ValueError("Invalid version format. See PEP 440 for details.")

    def __str__(self) -> str:
        major = self.major
        minor = self.minor
        if self.patch is None:
            patch = ""
        else:
            # notice the preceding dot (.)
            patch = f".{self.patch}"
        pre_release = self.pre_release or ""
        return f"{major}.{minor}{patch}{pre_release}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, VersionRepresentation):
            # we might compare None and [int, str]
            return all(
                (
                    self.major == other.major,
                    self.minor == other.minor,
                    self.patch == other.patch,
                    self.pre_release == other.pre_release,
                )
            )
        else:
            raise NotImplementedError

    def is_compatible(self, other: "VersionRepresentation") -> bool:
        return self.major == other.major and self.minor == other.minor

    def __lt__(self, other: "VersionRepresentation") -> bool:
        """Compare with a given VersionRepresentation

        .. warning::
            This method does NOT consider pre_release versions.
        """
        # check whether self < other
        lhs = (self.major, self.minor, self.patch or -1)
        rhs = (other.major, other.minor, other.patch or -1)
        return lhs < rhs


@dataclasses.dataclass(frozen=True)
class PythonVersion(VersionRepresentation):
    @property
    def full_representation(self) -> str:
        return f"Python{self.version}"

    @property
    def short_representation(self) -> str:
        return f"py{self.major}{self.minor}"

    @staticmethod
    def parse_short_representation(value: str) -> "PythonVersion":
        try:
            return _PythonVersions[value.upper()]
        except KeyError:
            raise KeyError(
                f"invalid value: {value}, must be one of {_PythonVersions.keys()}"
            ) from None


# NOTE(igarashi): PythonVersion class is immutable
_PythonVersions = {
    "PY27": PythonVersion(2, 7),
    "PY38": PythonVersion(3, 8),
    "PY39": PythonVersion(3, 9),
    "PY310": PythonVersion(3, 10),
    "PY311": PythonVersion(3, 11),
    "PY312": PythonVersion(3, 12),
}



import configparser
import pathlib
from collections import OrderedDict
from typing import Any, Dict, Optional

import tomlkit
import tomlkit.items
from tomlkit.toml_document import TOMLDocument

from .setting import SettingFile
from .types import PRIMITIVE_TYPES, SEQUENCE_TYPES


def dump_toml(path: pathlib.Path, setting: SettingFile) -> None:
    document: TOMLDocument = tomlkit.document()
    if path.exists():
        with path.open("r") as f:
            document = tomlkit.loads(f.read())

    # NOTE(igarashi): TOMLDocument inherits Dict
    entry = setting.entries()
    for p, d in sorted(entry):
        updated = SettingFile.update_by_entry(document, p, d)
        if isinstance(updated, tomlkit.items.Table):
            updated.comment(
                "# automatically generated by pysen\n"
                "# pysen ignores and overwrites any modifications"
            )

    with path.open("w") as f:
        buf = tomlkit.dumps(document)
        f.write(buf)


def _repr_cfg(data: Any) -> Optional[str]:
    if data is None:
        return None
    elif isinstance(data, PRIMITIVE_TYPES):
        return str(data)
    elif isinstance(data, SEQUENCE_TYPES):
        items = [_repr_cfg(x) for x in data]
        return ",".join(sorted(x for x in items if x is not None))
    else:
        raise RuntimeError(f"{type(data)} is not supported in cfg")


def dump_cfg(path: pathlib.Path, setting: SettingFile) -> None:
    flatten: Dict[str, Dict[str, Any]] = {}
    for p, d in setting.entries():
        if len(p) > 1:
            raise RuntimeError(
                "configparser cannot handle a section whose depth is more than 1"
            )
        flatten[p[0]] = d

    config = configparser.ConfigParser(allow_no_value=True)
    if path.exists():
        with path.open("r") as f:
            config.read_file(f)

    for p, d in sorted(flatten.items()):
        config[p] = OrderedDict()
        config.set(p, "# automatically generated by pysen", None)
        config.set(p, "# pysen ignores and overwrites any modifications", None)
        config[p].update(
            ((k, _repr_cfg(v)) for k, v in sorted(d.items()))  # type: ignore[misc]
        )

    with path.open("w") as f:
        config.write(f)


def dump(base_dir: pathlib.Path, fname: str, data: SettingFile) -> None:
    target_path = base_dir / fname
    ext = target_path.suffix
    if ext == ".toml":
        return dump_toml(target_path, data)
    elif ext == ".cfg" or ext == ".ini":
        return dump_cfg(target_path, data)
    else:
        raise RuntimeError(f"unknown extension: {ext}")



import logging
import pathlib
from typing import Callable, List, Optional, Sequence, Set, TypeVar

import tomlkit

from .component import ComponentBase
from .exceptions import InvalidConfigurationError
from .factory import configure_lint
from .manifest import ManifestBase
from .manifest_builder import build
from .mypy import Mypy
from .plugin_loader import load_plugin
from .pyproject_model import Config, LintConfig, has_tool_section, parse

_logger = logging.getLogger(__name__)
TConfig = TypeVar("TConfig")
# NOTE: The following names are in config resolution order
CONFIG_FILE_NAMES = ["pysen.toml", "pyproject.toml"]


def resolve_inheritance(
    path: pathlib.Path,
    selector: Callable[[pathlib.Path, Config], TConfig],
    base_selector: Callable[[pathlib.Path, TConfig], Optional[pathlib.Path]],
    updater: Callable[[TConfig, TConfig], TConfig],
    visited: Optional[Set[pathlib.Path]] = None,
) -> TConfig:
    visited = visited or set()
    if path in visited:
        raise InvalidConfigurationError(
            f"Circular dependency detected. {path} was visited more than once."
        )
    visited.add(path)

    config = parse(path)
    if config is None:
        raise InvalidConfigurationError(
            f"invalid base config: {path} doesn't exists or invalid file"
        )

    section = selector(path, config)
    base_path = base_selector(path, section)

    if base_path is None:
        return section

    base_section = resolve_inheritance(
        base_path, selector, base_selector, updater, visited
    )
    return updater(base_section, section)


def resolve_lint_config_inheritance(config: LintConfig) -> LintConfig:
    def selector(path: pathlib.Path, root: Config) -> LintConfig:
        if root.lint is None:
            raise InvalidConfigurationError(
                f"detected {path} doesn't have [tool.pysen.lint] section."
            )

        return root.lint

    def base_selector(path: pathlib.Path, c: LintConfig) -> Optional[pathlib.Path]:
        return c.base

    def updater(lhs: LintConfig, rhs: LintConfig) -> LintConfig:
        lhs.update(rhs)
        return lhs

    if config.base is None:
        return config

    base = resolve_inheritance(config.base, selector, base_selector, updater)
    base.update(config)
    return base


def load_lint_components(config: LintConfig) -> Sequence[ComponentBase]:
    config = resolve_lint_config_inheritance(config)
    components = configure_lint(config)

    # NOTE(igarashi): set convert_abspath=True to create a relative path from a absolute path
    if config.enable_mypy:
        mypy = next(c for c in components if isinstance(c, Mypy))
        mypy.setting._pysen_convert_abspath = True

        for _, s in mypy.module_settings.items():
            s._pysen_convert_abspath = True

    return components


def load_manifest(path: pathlib.Path) -> ManifestBase:
    config = parse(path)

    external_builder: Optional[pathlib.Path] = config.builder
    components: List[ComponentBase] = []

    if config.lint is not None:
        components.extend(load_lint_components(config.lint))

    if config.plugin is not None:
        for p in config.plugin:
            plugin = load_plugin(function=p.function, script=p.script)
            components.extend(plugin.load(path, p, config))

    return build(components, path, external_builder)


def _check_section_exists(config_path: pathlib.Path) -> bool:
    pyproject = tomlkit.loads(config_path.read_text())
    return has_tool_section("jiro", pyproject) or has_tool_section("pysen", pyproject)


def find_config(target_dir: pathlib.Path) -> Optional[pathlib.Path]:
    # NOTE: This method doesn't ensure if a config is valid for the configuration model
    for name in CONFIG_FILE_NAMES:
        path = target_dir / name
        if path.exists() and path.is_file():
            if _check_section_exists(path):
                _logger.debug(f"successfully found config file: {path}")
                return path

            _logger.debug(
                f"found a file, but [tool.pysen] section doesn't exist: {path}"
            )

    return None


def find_config_recursive(base_dir: pathlib.Path) -> Optional[pathlib.Path]:
    current = base_dir.resolve()
    while True:
        config = find_config(current)
        if config is not None:
            return config

        # reached root
        if current.parent == current:
            return None

        current = current.parent


def find_pyproject(path: Optional[pathlib.Path] = None) -> pathlib.Path:
    if path is not None:
        path = path.resolve()
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")
        elif not path.is_file():
            raise FileNotFoundError(f"{path} is not a file")

        return path
    else:
        cwd = pathlib.Path.cwd()
        p = find_config_recursive(cwd)
        if p is None:
            raise FileNotFoundError(
                "Could not find either a pyproject.toml file or "
                "a pysen.toml file containing a [tool.pysen] section "
                "in this or any of its parent directories. \n"
                "The `--loglevel debug option` may help."
            )
        return p



import argparse
import collections
import dataclasses
import pathlib
from abc import ABC, abstractmethod
from typing import Callable, DefaultDict, Dict, Iterable, List, Optional, Sequence

from .command import CommandBase
from .component import ComponentBase
from .dumper import dump
from .exceptions import InvalidComponentName
from .runner_options import PathContext, RunOptions
from .setting import SettingFile
from .types import ComponentName, TargetName

DumpHandlerType = Callable[
    [pathlib.Path, str, SettingFile],
    None,
]

ParserType = argparse._ActionsContainer
TargetType = List[CommandBase]


def get_targets(
    components: Sequence[ComponentBase],
) -> Dict[TargetName, List[ComponentName]]:
    result: DefaultDict[str, List[ComponentName]] = collections.defaultdict(list)
    for c in components:
        targets = c.targets
        for t in targets:
            result[t].append(c.name or "(no name)")

    return dict(result)


def get_target(
    target: TargetName,
    components: Sequence[ComponentBase],
    paths: PathContext,
    options: RunOptions,
) -> TargetType:
    result: TargetType = []
    for c in components:
        targets = c.targets
        if target in targets:
            result.append(c.create_command(target, paths, options))

    return result


def export_settings(
    paths: PathContext,
    components: Sequence[ComponentBase],
    dump_handler: DumpHandlerType,
) -> None:
    files: DefaultDict[str, SettingFile] = collections.defaultdict(SettingFile)
    for c in components:
        c.export_settings(paths, files)

    for fname, setting in files.items():
        try:
            dump_handler(
                paths.settings_dir,
                fname,
                setting,
            )
        except Exception as err:
            raise RuntimeError(f"got an unexpected error while creating {fname}: {err}")


class ManifestBase(ABC):
    def configure_parser(self, parser: ParserType) -> None:
        pass

    @abstractmethod
    def export_settings(self, paths: PathContext, args: argparse.Namespace) -> None:
        ...

    @abstractmethod
    def get_targets(
        self, args: argparse.Namespace
    ) -> Dict[TargetName, List[ComponentName]]:
        ...

    @abstractmethod
    def get_target(
        self,
        target: TargetName,
        paths: PathContext,
        options: RunOptions,
        args: argparse.Namespace,
    ) -> TargetType:
        ...


@dataclasses.dataclass(frozen=True)
class ManifestOptions:
    enabled: Optional[List[str]]
    disabled: Optional[List[str]]

    @staticmethod
    def configure_parser(parser: ParserType) -> None:
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--enable", action="append", help="component names to enable"
        )
        group.add_argument(
            "--disable", action="append", help="component names to disable"
        )

    @classmethod
    def from_parsed_arguments(cls, args: argparse.Namespace) -> "ManifestOptions":
        return cls(enabled=args.enable, disabled=args.disable)


def _verify_names(
    components: List[ComponentBase], names: List[str], option: str
) -> None:
    diff = set(names).difference(component.name for component in components)
    if diff:
        diff_str = ",".join(sorted(diff))
        raise InvalidComponentName(
            f"The following component(s) in option {option} were not found: {diff_str}"
        )


class Manifest(ManifestBase):
    def __init__(
        self,
        components: Optional[Sequence[ComponentBase]] = None,
        dump_handler: Optional[DumpHandlerType] = None,
    ) -> None:
        self._components: List[ComponentBase] = []
        if components is not None:
            self._components = list(components)

        self._dump_handler: DumpHandlerType = dump_handler or dump

    def configure_parser(self, parser: ParserType) -> None:
        ManifestOptions.configure_parser(parser)

    @property
    def components(self) -> List[ComponentBase]:
        return self._components

    def get_component(self, name: str) -> ComponentBase:
        for c in self._components:
            if c.name == name:
                return c
        else:
            raise KeyError(name)

    def _filter_components(self, args: argparse.Namespace) -> Iterable[ComponentBase]:
        options = ManifestOptions.from_parsed_arguments(args)
        enable = options.enabled
        disable = options.disabled

        if enable is None and disable is None:
            yield from self._components
            return

        predicate: Callable[[ComponentBase], bool]
        if enable is not None:
            _verify_names(self._components, enable, "--enable")

            def enable_predicate(x: ComponentBase) -> bool:
                assert enable is not None
                if x.name is None:
                    return False
                return x.name in enable

            predicate = enable_predicate
        elif disable is not None:
            _verify_names(self._components, disable, "--disable")

            def disable_predicate(x: ComponentBase) -> bool:
                assert disable is not None
                if x.name is None:
                    return True
                return x.name not in disable

            predicate = disable_predicate
        else:
            # NOTE(igarashi): `enable` and `disable` are in a mutually exclusive group
            raise AssertionError()

        for x in self._components:
            if predicate(x):
                yield x

    def export_settings(self, paths: PathContext, args: argparse.Namespace) -> None:
        # NOTE(igarashi): tomlkit doesn't keep the comments in the original file.
        # It will cause unnecessary changes when pysen doesn't export all configurations
        # since some comments like "automatically generated by pysen" would be partially removed
        # Therefore, ignore given args (`--disable` and `--enable`) as a workaround.
        export_settings(paths, self._components, self._dump_handler)

    def get_targets(
        self, args: argparse.Namespace
    ) -> Dict[TargetName, List[ComponentName]]:
        components = list(self._filter_components(args))
        return get_targets(components)

    def get_target(
        self,
        target: TargetName,
        paths: PathContext,
        options: RunOptions,
        args: argparse.Namespace,
    ) -> TargetType:
        components = list(self._filter_components(args))
        return get_target(target, components, paths, options)



import dataclasses
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional


@dataclasses.dataclass
class Diagnostic:
    file_path: Path
    start_line: Optional[int] = None
    end_line: Optional[int] = None
    start_column: Optional[int] = None
    message: Optional[str] = None
    diff: Optional[str] = None

    def __post_init__(self) -> None:
        assert self.file_path.is_absolute()
        if self.message is None and self.diff is None:
            raise ValueError("either message or diff must be filled")


class DiagnosticFormatter(ABC):
    @abstractmethod
    def format(self, diagnostic: Diagnostic, command_name: str) -> str:
        ...


def _format_diagnostic_position(diagnostic: Diagnostic) -> str:
    file_path = diagnostic.file_path.resolve()
    line = diagnostic.start_line or 1
    col = diagnostic.start_column or 1
    return f"{file_path}:{line}:{col}"


class _FLCMFormatter(DiagnosticFormatter):
    """
    FLCM: File, Line, Column, Message
    also see: https://www.gnu.org/prep/standards/html_node/Errors.html
    """

    def format(self, diagnostic: Diagnostic, command_name: str) -> str:
        position = _format_diagnostic_position(diagnostic)
        message: str
        if diagnostic.message:
            message = diagnostic.message.replace("\n", "\\n")
        elif diagnostic.diff:
            message = diagnostic.diff.replace("\n", "\\n")
        else:
            raise AssertionError()

        return f"{position}:{command_name}: {message}"


class _PrettyFormatter(DiagnosticFormatter):
    def format(self, diagnostic: Diagnostic, command_name: str) -> str:
        position = _format_diagnostic_position(diagnostic)
        message: str
        if diagnostic.message:
            message = diagnostic.message
        elif diagnostic.diff:
            message = f"\n{diagnostic.diff}"
        else:
            raise AssertionError()

        return f"[{command_name}] {position}:{message}"


FLCMFormatter = _FLCMFormatter()
PrettyFormatter = _PrettyFormatter()



import pathlib
from typing import DefaultDict, Optional, Sequence

from pysen.ext import black_wrapper
from pysen.ext.black_wrapper import BlackSetting

from .command import CommandBase
from .component import LintComponentBase
from .lint_command import LintCommandBase
from .path import resolve_path
from .reporter import Reporter
from .runner_options import PathContext, RunOptions
from .setting import SettingFile
from .source import PythonFileFilter, Source

_SettingFileName = "pyproject.toml"


class BlackCommand(LintCommandBase):
    def __init__(
        self,
        name: str,
        paths: PathContext,
        source: Source,
        inplace_edit: bool,
    ) -> None:
        super().__init__(paths.base_dir, source)
        self._name = name
        self._setting_path = resolve_path(paths.settings_dir, _SettingFileName)
        self._inplace_edit = inplace_edit

    @property
    def name(self) -> str:
        return self._name

    @property
    def has_side_effects(self) -> bool:
        return self._inplace_edit

    def __call__(self, reporter: Reporter) -> int:
        sources = self._get_sources(reporter, PythonFileFilter)
        reporter.logger.info(f"Checking {len(sources)} files")
        return black_wrapper.run(
            reporter, self.base_dir, self._setting_path, sources, self._inplace_edit
        )

    def run_files(self, reporter: Reporter, files: Sequence[pathlib.Path]) -> int:
        covered_files = self._get_covered_files(reporter, files, PythonFileFilter)

        if len(covered_files) == 0:
            return 0

        return black_wrapper.run(
            reporter,
            self.base_dir,
            self._setting_path,
            covered_files,
            self._inplace_edit,
        )


class Black(LintComponentBase):
    def __init__(
        self,
        name: str = "black",
        setting: Optional[BlackSetting] = None,
        source: Optional[Source] = None,
    ) -> None:
        super().__init__(name, source)

        self._setting = setting or BlackSetting.default()

    @property
    def setting(self) -> BlackSetting:
        return self._setting

    @setting.setter
    def setting(self, value: BlackSetting) -> None:
        self._setting = value

    def export_settings(
        self,
        paths: PathContext,
        files: DefaultDict[str, SettingFile],
    ) -> None:
        setting_file = files[_SettingFileName]
        section, setting = self._setting.export()
        setting_file.set_section(section, setting)

    @property
    def targets(self) -> Sequence[str]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        if target == "lint":
            return BlackCommand(self.name, paths, self.source, False)
        elif target == "format":
            return BlackCommand(self.name, paths, self.source, True)

        raise AssertionError(f"unknown {target}")



import collections
import contextlib
import dataclasses
import logging
import logging.handlers
from typing import DefaultDict, Iterator, List, Optional, cast

import colorlog
from colorlog import ColoredFormatter

_PROCESS_STDOUT_LEVEL = 11
_PROCESS_STDOUT = "STDOUT"
_PROCESS_STDERR_LEVEL = 12
_PROCESS_STDERR = "STDERR"

logging.addLevelName(_PROCESS_STDOUT_LEVEL, _PROCESS_STDOUT)
logging.addLevelName(_PROCESS_STDERR_LEVEL, _PROCESS_STDERR)


def _concat_logging_path(lhs: str, rhs: str) -> str:
    return f"{lhs}.{rhs}"


PYSEN_LOGGER_PREFIX = "__PYSEN__"
REPORTER_LOGGER_PREFIX = _concat_logging_path(PYSEN_LOGGER_PREFIX, "reporter")
PROCESS_OUTPUT_LOGGER_PREFIX = _concat_logging_path(PYSEN_LOGGER_PREFIX, "proc")

pysen_root_logger = logging.getLogger(PYSEN_LOGGER_PREFIX)
pysen_root_logger.propagate = False

reporter_root_logger = logging.getLogger(REPORTER_LOGGER_PREFIX)
reporter_root_logger.setLevel(logging.CRITICAL)

process_output_root_logger = logging.getLogger(PROCESS_OUTPUT_LOGGER_PREFIX)
process_output_root_logger.setLevel(logging.INFO)
# NOTE(igarashi): Set NullHandler() so that process_output_root_logger does not call
# the lastResort logger when propagate = False.
# See: https://github.com/python/cpython/blob/c3dd7e45cc5d36bbe2295c2840faabb5c75d83e4/Lib/logging/__init__.py#L1672-L1679  # NOQA
process_output_root_logger.addHandler(logging.NullHandler())
process_output_root_logger.propagate = False

_logging_output_colors = {
    **colorlog.default_log_colors,
    **{_PROCESS_STDOUT: "", _PROCESS_STDERR: "yellow"},
}


def get_reporter_logger(name: str) -> logging.Logger:
    return logging.getLogger(_concat_logging_path(REPORTER_LOGGER_PREFIX, name))


def get_process_output_logger(name: str) -> logging.Logger:
    return logging.getLogger(_concat_logging_path(PROCESS_OUTPUT_LOGGER_PREFIX, name))


NamedRecords = DefaultDict[str, List[logging.LogRecord]]


class _GroupedMemoryHandler(logging.handlers.MemoryHandler):
    def __init__(self, target: logging.Handler, capacity: int = 1024) -> None:
        super().__init__(capacity, target=target)
        self._named_records: NamedRecords = collections.defaultdict(list)

    def emit(self, record: logging.LogRecord) -> None:
        self._named_records[record.name.split(".")[-1]].append(record)

    def shouldFlush(self, record: logging.LogRecord) -> bool:
        return False

    def setFormatter(self, fmt: Optional[logging.Formatter]) -> None:
        self.target: Optional[logging.Handler]
        assert self.target is not None
        self.target.setFormatter(fmt)

    def flush(self) -> None:
        names = sorted(self._named_records.keys())
        for name in names:
            for record in self._named_records[name]:
                super().emit(record)
        super().flush()
        self._named_records = collections.defaultdict(list)


def setup_logger(loglevel: int, pretty: bool = True) -> None:
    root_logger = logging.getLogger("pysen")
    root_logger.setLevel(loglevel)
    root_logger.handlers.clear()

    handler = logging.StreamHandler()
    handler.setFormatter(_general_formatter(pretty))
    root_logger.addHandler(handler)


class _LoggingUnit:
    def __init__(
        self, loglevel: int, is_grouped: bool, pretty: bool, is_process_enabled: bool
    ) -> None:
        self._loglevel = loglevel
        self._is_grouped = is_grouped
        self._pretty = pretty
        self._is_process_enabled = is_process_enabled
        self._handler = self._create_handler()

    def setup(self) -> None:
        self._handler.setLevel(logging.DEBUG)
        self._handler.setFormatter(self._create_formatter())
        pysen_root_logger.addHandler(self._handler)
        reporter_root_logger.setLevel(self._loglevel)
        if self._is_process_enabled:
            process_output_root_logger.propagate = True

    def finalize(self) -> None:
        self._handler.flush()
        pysen_root_logger.removeHandler(self._handler)
        process_output_root_logger.propagate = False

    def _create_handler(self) -> logging.Handler:
        if self._is_grouped:
            return _GroupedMemoryHandler(target=logging.StreamHandler())
        else:
            return logging.StreamHandler()

    def _create_formatter(self) -> logging.Formatter:
        if self._pretty:
            ret: logging.Formatter = _CustomColoredFormatter(
                "%(log_color)s%(message)s",
                log_colors=_logging_output_colors,
            )
            return ret
        else:
            return logging.Formatter("%(message)s")


def _general_formatter(pretty: bool) -> logging.Formatter:
    if pretty:
        ret: logging.Formatter = ColoredFormatter("%(log_color)s%(message)s")
        return ret
    else:
        return logging.Formatter("%(message)s")


def _get_process_output_level_name(loglevel: int) -> str:
    if loglevel <= logging.INFO:
        return _PROCESS_STDOUT
    else:
        return _PROCESS_STDERR


class _CustomColoredFormatter(colorlog.ColoredFormatter):  # type: ignore
    def format(self, record: logging.LogRecord) -> str:
        if record.name.startswith(PROCESS_OUTPUT_LOGGER_PREFIX):
            record.levelname = _get_process_output_level_name(record.levelno)

        return cast(str, super().format(record))


@dataclasses.dataclass(frozen=True)
class CommandLoggingOptions:
    is_grouped: bool
    pretty: bool
    process_output: bool

    @contextlib.contextmanager
    def start_logging(self, loglevel: int) -> Iterator[None]:
        unit = _LoggingUnit(loglevel, self.is_grouped, self.pretty, self.process_output)
        unit.setup()

        try:
            yield
        finally:
            unit.finalize()



import dataclasses
import pathlib
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set

from .git_utils import GitRepositoryNotFoundError, check_tracked, list_indexed_files
from .path import PathLikeType, is_contained, resolve_path
from .reporter import Reporter

FilePredicateType = Callable[[pathlib.Path], bool]


def extension_filter(accept_extensions: Set[str]) -> FilePredicateType:
    def impl(path: pathlib.Path) -> bool:
        return path.suffix in accept_extensions

    return impl


PythonFileFilter = extension_filter({".py", ".pyi"})


@dataclasses.dataclass
class SourceEntrySetting:
    glob: bool
    base_dir: Optional[pathlib.Path] = None


def _resolve(
    base_dir: pathlib.Path, p: PathLikeType, setting: SourceEntrySetting
) -> List[pathlib.Path]:
    if isinstance(p, pathlib.Path) and setting.glob:
        raise RuntimeError("cannot use pathlib.Path when glob=True")

    if setting.base_dir is not None:
        base_dir = setting.base_dir
    base_dir = base_dir.resolve()

    if not setting.glob:
        return [resolve_path(base_dir, p)]

    assert isinstance(p, str)
    # NOTE(igarashi): ensure that p is an instance of str
    # otherwise, pathlib.glob raises an error when sys.intern() is called
    p = str(p)

    # NOTE(igarashi): pathlib.Path.glob(".") raises IndexError in Python 3.7 (might be a bug?)
    # To avoid the bug, add a condition to handle the case for the workaround.
    if p == ".":
        return [resolve_path(base_dir, ".")]

    # run base_dir.glob only if isinstance(p, str) and setting.glob
    return [resolve_path(base_dir, g) for g in base_dir.glob(p)]


class Source:
    def __init__(
        self,
        includes: Optional[Sequence[PathLikeType]] = None,
        excludes: Optional[Sequence[PathLikeType]] = None,
        include_globs: Optional[Sequence[PathLikeType]] = None,
        exclude_globs: Optional[Sequence[PathLikeType]] = None,
    ) -> None:
        self._includes: Dict[PathLikeType, SourceEntrySetting] = {}
        self._excludes: Dict[PathLikeType, SourceEntrySetting] = {}

        if includes is not None:
            for i in includes:
                self.add_include(i, glob=False)

        if include_globs is not None:
            for i in include_globs:
                self.add_include(i, glob=True)

        if excludes is not None:
            for e in excludes:
                self.add_exclude(e, glob=False)

        if exclude_globs is not None:
            for e in exclude_globs:
                self.add_exclude(e, glob=True)

    def add_include(
        self,
        entry: PathLikeType,
        *,
        glob: bool = False,
        base_dir: Optional[pathlib.Path] = None,
    ) -> None:
        self._includes[entry] = SourceEntrySetting(glob=glob, base_dir=base_dir)

    def add_exclude(
        self,
        entry: PathLikeType,
        *,
        glob: bool = False,
        base_dir: Optional[pathlib.Path] = None,
    ) -> None:
        self._excludes[entry] = SourceEntrySetting(glob=glob, base_dir=base_dir)

    def remove_include(self, entry: PathLikeType) -> None:
        self._includes.pop(entry)

    def remove_exclude(self, entry: PathLikeType) -> None:
        self._excludes.pop(entry)

    @property
    def includes(self) -> Dict[PathLikeType, SourceEntrySetting]:
        return self._includes

    @property
    def excludes(self) -> Dict[PathLikeType, SourceEntrySetting]:
        return self._excludes

    def copy(self) -> "Source":
        new = Source()
        for path, setting in self.includes.items():
            new.add_include(path, glob=setting.glob, base_dir=setting.base_dir)

        for path, setting in self.excludes.items():
            new.add_exclude(path, glob=setting.glob, base_dir=setting.base_dir)

        return new

    def iter_include_entries(self, base_dir: pathlib.Path) -> Iterable[pathlib.Path]:
        for include, setting in self._includes.items():
            yield from iter(_resolve(base_dir, include, setting))

    def iter_exclude_entries(self, base_dir: pathlib.Path) -> Iterable[pathlib.Path]:
        for exclude, setting in self._excludes.items():
            yield from iter(_resolve(base_dir, exclude, setting))

    def _resolve_include_files(
        self,
        base_dir: pathlib.Path,
        filter_predicate: FilePredicateType,
        use_git: bool,
        reporter: Optional[Reporter] = None,
    ) -> Set[pathlib.Path]:
        includes = self.iter_include_entries(base_dir)

        included_files: Set[pathlib.Path] = set()

        for include in includes:
            if not include.exists():
                continue
            if include.is_file():
                # NOTE(igarashi): include this file anyway even though it is
                # not a .py file (e.g., script/command)
                if use_git:
                    try:
                        if not check_tracked(include):
                            continue
                    except GitRepositoryNotFoundError:
                        if reporter is not None:
                            reporter.logger.warning(
                                f"{include} is outside repository. ignored."
                            )

                included_files.add(include)
            else:
                if use_git:
                    try:
                        included_files.update(
                            x
                            for x in list_indexed_files(include)
                            if filter_predicate(x)
                        )
                    except GitRepositoryNotFoundError:
                        if reporter is not None:
                            reporter.logger.warning(
                                f"{include} is outside repository. ignored."
                            )
                else:
                    included_files.update(
                        x for x in include.glob("**/*") if filter_predicate(x)
                    )

        return included_files

    def resolve_files(
        self,
        base_dir: pathlib.Path,
        filter_predicate: Optional[FilePredicateType],
        use_git: bool = True,
        reporter: Optional[Reporter] = None,
    ) -> Set[pathlib.Path]:
        """Returns a set of target files.
        Note:
            `filter_predicate` is used to filter files only when a directory is expanded by
            the system. If an user explicitly specifies the path as included, it will be added
            to the return object.

        Examples:
            >>> s = Source(includes=["foo/", "bar/doc.md"])
            >>> s.resolve_files(base_dir, lambda x: x.suffix == ".py")
            [PosixPath("foo/main.py"), PosixPath("foo/module.py"), PosixPath("bar/doc.md")]
        """

        def _default(x: pathlib.Path) -> bool:
            return True

        filter_predicate = filter_predicate or _default

        assert base_dir.is_absolute()

        included = self._resolve_include_files(
            base_dir, filter_predicate, use_git, reporter
        )

        exclude_entries = list(self.iter_exclude_entries(base_dir))

        def exclude_filter(path: pathlib.Path) -> bool:
            return not any(is_contained(exclude, path) for exclude in exclude_entries)

        return set(filter(exclude_filter, included))



import itertools
import pathlib
from enum import Enum
from typing import Any, Callable, DefaultDict, Dict, List, Mapping, Optional, Sequence

from .command import CommandBase
from .component import ComponentBase
from .ext import mypy_wrapper
from .ext.mypy_wrapper import (  # NOQA
    MypyFollowImports,
    MypyPlugin,
    MypySetting,
    MypyTarget,
)
from .path import is_covered, resolve_path
from .reporter import Reporter
from .runner_options import PathContext, RunOptions
from .setting import SettingFile

_SettingFileName = "setup.cfg"


def _get_differences_from_base(
    entries: Dict[str, Any], base_entries: Dict[str, Any]
) -> Dict[str, Any]:
    # NOTE(igarashi): do not use `dict(entries.items() - base_entries.items())`
    # since entries may contain an unhashable type like lists.
    added = entries.keys() - base_entries.keys()
    duplicated = entries.keys() & base_entries.keys()

    diff: Dict[str, Any] = {}

    for key in added:
        diff[key] = entries[key]

    for key in duplicated:
        if entries[key] == base_entries[key]:
            continue

        diff[key] = entries[key]

    # NOTE(igarashi): Since there is no way to obtain a default value for each removed item,
    # we cannot put them to diff.
    # TODO(igarashi): Implement field attribute to each MypySetting option so that
    # we can get a default value. Uncomment the following code:
    # _IgnoreDifferenceFields = {"_pysen_convert_abspath", "python_version", "mypy_path"}
    # removed = base_entries.keys() - entries.keys() - _IgnoreDifferenceFields

    return diff


class MypyPreset(Enum):
    VERY_STRICT = (MypySetting.very_strict,)
    STRICT = (MypySetting.strict,)
    ENTRY = (MypySetting.entry,)

    def __init__(self, factory: Callable[..., MypySetting]) -> None:
        self._factory = factory

    def get_setting(self, **kwargs: Any) -> MypySetting:
        return self._factory(**kwargs)


class MypyCommand(CommandBase):
    def __init__(
        self,
        name: str,
        paths: PathContext,
        mypy_targets: Sequence[MypyTarget],
        require_diagnostics: bool,
    ) -> None:
        self._name = name
        self._base_dir = paths.base_dir
        self._mypy_targets: List[MypyTarget] = list(mypy_targets)

        self._setting_path = resolve_path(paths.settings_dir, _SettingFileName)
        self._require_diagnostics = require_diagnostics

    @property
    def name(self) -> str:
        return self._name

    @property
    def has_side_effects(self) -> bool:
        return False

    @property
    def base_dir(self) -> pathlib.Path:
        return self._base_dir

    @property
    def setting_path(self) -> pathlib.Path:
        return self._setting_path

    def __call__(self, reporter: Reporter) -> int:
        exit_code: int = 0
        num_targets = len(self._mypy_targets)

        if num_targets == 0:
            reporter.logger.error(
                "No mypy targets specified. "
                "You must specify at least one entry in `tool.pysen.lint.mypy_targets`."
            )
            return 2

        for idx, target in enumerate(self._mypy_targets):
            reporter.logger.info(
                f"[{idx+1}/{num_targets}] Checking {len(target.paths)} entries"
            )
            ret = mypy_wrapper.run(
                reporter,
                self.base_dir,
                self.setting_path,
                target,
                self._require_diagnostics,
            )
            if ret != 0:
                exit_code = ret

        return exit_code

    def run_files(self, reporter: Reporter, files: Sequence[pathlib.Path]) -> int:
        sources = list(
            itertools.chain.from_iterable(target.paths for target in self._mypy_targets)
        )
        covered_files: List[pathlib.Path] = []
        for f in files:
            if is_covered(f, sources):
                covered_files.append(f)
            else:
                reporter.logger.info(f"Skipping {f} for {self._name}")

        if len(covered_files) == 0:
            return 0

        return mypy_wrapper.run(
            reporter,
            self.base_dir,
            self._setting_path,
            MypyTarget(covered_files),
            self._require_diagnostics,
        )


class Mypy(ComponentBase):
    def __init__(
        self,
        name: str = "mypy",
        mypy_targets: Optional[Sequence[MypyTarget]] = None,
        setting: Optional[MypySetting] = None,
        module_settings: Optional[Mapping[str, MypySetting]] = None,
    ) -> None:
        self._name = name
        self._mypy_targets = list(mypy_targets or [])
        self._setting: MypySetting = setting or MypySetting()
        self._module_settings: Dict[str, MypySetting] = dict(module_settings or {})

    @property
    def name(self) -> str:
        return self._name

    @property
    def setting(self) -> MypySetting:
        return self._setting

    @property
    def module_settings(self) -> Dict[str, MypySetting]:
        return self._module_settings

    @property
    def mypy_targets(self) -> List[MypyTarget]:
        return self._mypy_targets

    def export_settings(
        self,
        paths: PathContext,
        files: DefaultDict[str, SettingFile],
    ) -> None:
        setting_file = files[_SettingFileName]
        global_section, global_setting = self._setting.export(paths.base_dir)
        setting_file.set_section(global_section, global_setting)

        for module_name, setting in self._module_settings.items():
            section, module_setting = setting.export(
                paths.base_dir, target_module=module_name
            )
            module_setting = _get_differences_from_base(module_setting, global_setting)
            setting_file.set_section(section, module_setting)

    @property
    def targets(self) -> Sequence[str]:
        return ["lint"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        if target == "lint":
            return MypyCommand(
                self.name, paths, self.mypy_targets, options.require_diagnostics
            )

        raise AssertionError(f"unknown {target}")



import logging
import pathlib
from typing import Any, List, Optional, Sequence

import pysen.pyproject  # NOQA

from ._version import __version__  # NOQA
from .black import Black, BlackSetting  # NOQA
from .command import CommandBase  # NOQA
from .component import ComponentBase  # NOQA
from .factory import ConfigureLintOptions, configure_lint  # NOQA
from .flake8 import Flake8, Flake8Setting  # NOQA
from .isort import Isort, IsortSetting  # NOQA
from .lint_command import SingleFileFormatCommandBase, SingleFileLintCommandBase  # NOQA
from .logging_utils import setup_logger  # NOQA
from .manifest import Manifest, ManifestBase  # NOQA
from .manifest_builder import build as build_manifest  # NOQA
from .mypy import Mypy, MypyPreset, MypySetting, MypyTarget  # NOQA
from .path import PathLikeType  # NOQA
from .plugin import PluginBase, PluginConfig  # NOQA
from .py_version import PythonVersion  # NOQA
from .pyproject import load_manifest  # NOQA
from .pyproject_model import Config  # NOQA
from .reporter import ReporterFactory  # NOQA
from .runner import Runner  # NOQA
from .runner_options import RunOptions  # NOQA
from .source import Source  # NOQA

_logger = logging.getLogger(__name__)


try:
    from .setuptools import generate_setting_files  # NOQA isort:skip
    from .setuptools import setup  # NOQA isort:skip
    from .setuptools import setup_from_pyproject  # NOQA isort:skip
except ImportError:
    _logger.warning(
        "[pysen.setuptools] failed to import setuptools.\n"
        "You won't be able to invoke pysen via setuptools.\n"
        "If you use pysen CLI directly (i.e., `pysen run ...`), you may ignore this warning."
    )


def run(
    base_dir: pathlib.Path,
    target_name: str,
    manifest_args: Optional[Sequence[str]] = None,
    reporter_factory: Optional[ReporterFactory] = None,
    *,
    settings_dir: Optional[pathlib.Path] = None,
    options: Optional[RunOptions] = None,
    pyproject: Optional[pathlib.Path] = None,
    components: Optional[Sequence[ComponentBase]] = None,
    manifest: Optional[ManifestBase] = None,
) -> bool:
    func_args: List[Any] = [x is not None for x in (pyproject, components, manifest)]
    if func_args.count(True) != 1:
        raise ValueError(
            "only one of pyproject, components, and manifest must be specified"
        )

    target: ManifestBase
    if manifest is not None:
        target = manifest
    elif components is not None:
        target = build_manifest(list(components))
    else:
        assert pyproject is not None
        target = load_manifest(pyproject)

    manifest_args = manifest_args or []
    reporter_factory = reporter_factory or ReporterFactory()

    assert manifest_args is not None
    assert reporter_factory is not None

    runner = Runner(target)
    parsed_args = runner.parse_manifest_arguments(manifest_args)
    options = options or RunOptions()
    runner.run(
        target_name,
        base_dir,
        parsed_args,
        reporter_factory,
        options,
        files=None,
        settings_dir=settings_dir,
    )

    return not reporter_factory.has_error()



import contextlib
import os
import pathlib
from typing import Iterable, Iterator, Union

PathLikeType = Union[pathlib.Path, str]


def wrap_path(s: PathLikeType) -> pathlib.Path:
    if isinstance(s, pathlib.Path):
        return s
    return pathlib.Path(s)


def resolve_path(base_dir: pathlib.Path, path: PathLikeType) -> pathlib.Path:
    return (base_dir / path).expanduser().resolve()


def get_relative_path(path: PathLikeType, base_dir: pathlib.Path) -> str:
    pth = wrap_path(path)
    if not pth.is_absolute():
        return str(pth)

    return os.path.relpath(pth, base_dir)


@contextlib.contextmanager
def change_dir(dst: pathlib.Path) -> Iterator[None]:
    old = pathlib.Path.cwd()
    try:
        os.chdir(dst)
        yield
    finally:
        os.chdir(old)


def is_covered(path: pathlib.Path, sources: Iterable[pathlib.Path]) -> bool:
    """
    Checks if `path` is contained in any of the subdirectories in sources.
    See the test cases for details.
    """
    path = path.resolve()
    abs_path = [source.resolve() for source in sources]
    return any(c in abs_path for c in list(path.parents) + [path])


def is_contained(parent: pathlib.Path, child: pathlib.Path) -> bool:
    if not parent.is_absolute() or not child.is_absolute():
        raise ValueError("Argument 'parent' and 'child' must be absolute")
    return str(child).startswith(str(parent))



import pathlib
import subprocess
from abc import ABC, abstractmethod
from typing import Sequence

from .exceptions import CommandNotFoundError, RunTargetFileNotSupported
from .reporter import Reporter


class CommandBase(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def __call__(self, reporter: Reporter) -> int:
        ...

    @property
    def has_side_effects(self) -> bool:
        return True

    def run(self, reporter: Reporter) -> int:
        return self.__call__(reporter)

    def run_files(self, reporter: Reporter, files: Sequence[pathlib.Path]) -> int:
        raise RunTargetFileNotSupported(self.name)


def check_command_installed(
    *validation_command: str,
) -> None:
    err = CommandNotFoundError(
        f"The command `{' '.join(validation_command)}` failed."
        " Make sure it is installed."
    )
    try:
        retval = subprocess.call(
            validation_command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        # This will be raised when self.validation_command[0] does not exist.
        raise err
    if retval == 127:
        # In some cases (e.g. pyenv), FileNotFoundError is not raised.
        # Instead, we look at the return code to tell if the command could not be found.
        raise err
